from typing import Any, Dict, List, Optional, Union, cast

from sqlalchemy import Column, Table, func, select
from sqlalchemy.sql import Select, and_
from sqlalchemy.sql.dml import Delete, Insert, Update

from ..db.container import Database
from ..types import Connection, Record, Records

QueryType = Union[Delete, Update, Select]


class CrudDB(Database):
    """A :class:`.Database` with additional methods for CRUD operations"""

    async def db_select(
        self,
        table: Table,
        filters: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Records:
        """Select rows from a given table

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        query = self.get_query(table, table.select(), consumer=consumer, params=filters)
        async with self.ensure_connection(conn) as conn:
            return await conn.execute(query)

    async def db_delete(
        self,
        table: Table,
        filters: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Records:
        """Delete rows from a given table

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        query = self.get_query(
            table,
            table.delete().returning(*table.columns),
            consumer=consumer,
            params=filters,
        )
        async with self.ensure_connection(conn) as conn:
            return await conn.execute(query)

    async def db_count(
        self,
        table: Table,
        filters: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> int:
        """Count rows in a table

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        query = self.get_query(table, table.select(), consumer=consumer, params=filters)
        return await self.db_count_query(query, conn=conn)

    async def db_count_query(self, query, *, conn: Optional[Connection] = None) -> int:
        count_query = select([func.count()]).select_from(query.alias("inner"))
        async with self.ensure_connection(conn) as conn:
            result = await conn.execute(count_query)
            return result.scalar()

    async def db_insert(
        self,
        table: Table,
        data: Union[List[Dict], Dict],
        *,
        conn: Optional[Connection] = None,
    ) -> Records:
        """Perform an insert into a table

        :param table: sqlalchemy Table
        :param data: key-value pairs for columns values
        :param conn: optional db connection
        """
        async with self.ensure_connection(conn) as conn:
            query = self.get_insert(table, data)
            return await conn.execute(query)

    async def db_update(
        self,
        table: Table,
        filters: Dict,
        data: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Records:
        """Perform an update of rows

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows to update
        :param data: key-value pairs for updating columns values of selected rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        update = (
            cast(
                Update,
                self.get_query(
                    table, table.update(), consumer=consumer, params=filters
                ),
            )
            .values(**data)
            .returning(*table.columns)
        )
        async with self.ensure_connection(conn) as conn:
            return await conn.execute(update)

    async def db_upsert(
        self,
        table: Table,
        filters: Dict,
        data: Optional[Dict] = None,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Record:
        """Perform an upsert for a single record

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows to update
        :param data: key-value pairs for updating columns values of selected rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        if data:
            result = await self.db_update(
                table, filters, data, conn=conn, consumer=consumer
            )
        else:
            result = await self.db_select(table, filters, conn=conn, consumer=consumer)
        record = result.one_or_none()
        if record is None:
            insert_data = data.copy() if data else {}
            insert_data.update(filters)
            result = await self.db_insert(table, insert_data, conn=conn)
            record = result.one()
        return record

    def get_insert(self, table: Table, records: Union[List[Dict], Dict]) -> Insert:
        if isinstance(records, dict):
            records = [records]
        return table.insert(records).returning(*table.columns)

    def get_query(
        self,
        table: Table,
        query: QueryType,
        *,
        params: Optional[Dict] = None,
        consumer: Any = None,
    ) -> QueryType:
        """Build an SqlAlchemy query

        :param table: sqlalchemy Table
        :param query: sqlalchemy query type
        :param params: key-value pairs for the query
        :param consumer: optional consumer for manipulating parameters
        """
        filters: List = []
        columns = table.c
        params = params or {}

        for key, value in params.items():
            bits = key.split(":")
            field = bits[0]
            op = bits[1] if len(bits) == 2 else "eq"
            filter_field = getattr(consumer, f"filter_{field}", None)
            if filter_field:
                result = filter_field(op, value)
            else:
                field = getattr(columns, field)
                result = self.default_filter_field(field, op, value)
            if result is not None:
                if not isinstance(result, (list, tuple)):
                    result = (result,)
                filters.extend(result)
        if filters:
            whereclause = and_(*filters) if len(filters) > 1 else filters[0]
            query = cast(Select, query).where(whereclause)
        return query

    def default_filter_field(self, field: Column, op: str, value: Any):
        """
        Applies a filter on a field.

        Notes on 'ne' op:

        Example data: [None, 'john', 'roger']
        ne:john would return only roger (i.e. nulls excluded)
        ne:     would return john and roger

        Notes on  'search' op:

        For some reason, SQLAlchemy uses to_tsquery rather than
        plainto_tsquery for the match operator

        to_tsquery uses operators (&, |, ! etc.) while
        plainto_tsquery tokenises the input string and uses AND between
        tokens, hence plainto_tsquery is what we want here

        For other database back ends, the behaviour of the match
        operator is completely different - see:
        http://docs.sqlalchemy.org/en/rel_1_0/core/sqlelement.html

        :param field: field name
        :param op: 'eq', 'ne', 'gt', 'lt', 'ge', 'le' or 'search'
        :param value: comparison value, string or list/tuple
        :return:
        """
        multiple = isinstance(value, (list, tuple))

        if value == "":
            value = None

        if multiple and op in ("eq", "ne"):
            if op == "eq":
                return field.in_(value)
            elif op == "ne":
                return ~field.in_(value)
        else:
            if multiple:
                assert len(value) > 0
                value = value[0]

            if op == "eq":
                return field == value
            elif op == "ne":
                return field != value
            elif op == "gt":
                return field > value
            elif op == "ge":
                return field >= value
            elif op == "lt":
                return field < value
            elif op == "le":
                return field <= value



"""Testing utilities
"""
import asyncio
from contextlib import asynccontextmanager, contextmanager
from typing import Any

from aiohttp.client import ClientResponse

from .db import CrudDB, Database
from .json import dumps, loads
from .types import Connection


async def json_body(response: ClientResponse, status: int = 200) -> Any:
    assert response.content_type == "application/json"
    data = await response.json(loads=loads)
    if response.status != status:  # pragma: no cover
        print(dumps({"status": response.status, "data": data}, indent=4))

    assert response.status == status
    return data


@contextmanager
def with_test_db(db: CrudDB) -> CrudDB:
    db.create_all()
    try:
        yield db
    finally:
        db.drop_all_schemas()


class SingleConnDatabase(CrudDB):
    """Useful for speedup testing"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._lock = asyncio.Lock()
        self._connection = None

    @classmethod
    def from_db(cls, db: Database) -> "SingleConnDatabase":
        return cls(dsn=db.dsn, metadata=db.metadata)

    async def __aenter__(self) -> "SingleConnDatabase":
        self._connection = await self.engine.begin()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        transaction = self._connection.get_transaction()
        await transaction.rollback()
        self._connection = None

    @asynccontextmanager
    async def connection(self) -> Connection:
        async with self._lock:
            yield self._connection

    @asynccontextmanager
    async def transaction(self) -> Connection:
        async with self._lock:
            yield self._connection



from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence

from aiohttp.web import Application

from .cli import OpenApiClient
from .data.fields import Choice, bool_field, integer_field, str_field
from .data.pagination import MAX_PAGINATION_LIMIT
from .spec import OpenApi, OpenApiSpec, Redoc
from .spec.utils import docjoin


def rest(
    openapi: Optional[Dict] = None,
    setup_app: Callable[[Application], None] = None,
    base_path: str = "",
    commands: Optional[List] = None,
    allowed_tags: Sequence[str] = (),
    validate_docs: bool = False,
    servers: Optional[List[str]] = None,
    security: Optional[Dict[str, Dict]] = None,
    redoc: Optional[Redoc] = None,
    OpenApiSpecClass: type = OpenApiSpec,
    **kwargs,
) -> OpenApiClient:
    """Create the OpenApi Client"""
    if openapi is not None:
        openapi = OpenApiSpecClass(
            OpenApi(**(openapi or {})),
            allowed_tags=allowed_tags,
            validate_docs=validate_docs,
            servers=servers,
            security=security,
            redoc=redoc,
        )
    return OpenApiClient(
        spec=openapi,
        base_path=base_path,
        commands=commands,
        setup_app=setup_app,
        **kwargs,
    )


@dataclass
class Query:
    """Base dataclass for querying pagination"""

    limit: int = integer_field(
        min_value=1,
        max_value=MAX_PAGINATION_LIMIT,
        description="Limit the number of objects returned from the endpoint",
    )
    offset: int = integer_field(
        min_value=0,
        description=(
            "Number of objects to exclude. "
            "Use in conjunction with limit to paginate results"
        ),
    )


def orderable(*orderable_fields) -> type:
    """Create a dataclass with `order_by` choice field and the `order_desc`
    boolean field.

    :param *orderable_fields: fields which can be used for ordering
    """

    @dataclass
    class Orderable:
        order_by: str = str_field(
            validator=Choice(orderable_fields),
            description=(
                "Order results by given column (default ascending order). "
                f"Possible values are {docjoin(orderable_fields)}"
            ),
        )
        order_desc: bool = bool_field(
            description=("Change order direction to descending")
        )

    return Orderable


def searchable(*searchable_fields) -> type:
    """Create a dataclass with `search_fields` class attribute and `search` field.
    The search field is a set of field which can be used for searching and it is used
    internally by the library, while the `search` field is the query string passed
    in the url.

    :param searchable_fields: fields which can be used for searching
    """
    fields = docjoin(searchable_fields)

    @dataclass
    class Searchable:
        search_fields = frozenset(searchable_fields)
        search: str = str_field(
            description=(
                "Search query string. " f"The search is performed on {fields} fields."
            )
        )

    return Searchable



import re
from typing import Dict, List, Optional, Sequence, cast

import sqlalchemy as sa
from aiohttp import web
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import Select, or_

from ..data.pagination import PaginatedData, Pagination
from ..db.dbmodel import CrudDB
from ..spec.path import ApiPath
from ..types import Connection, DataType, Record, Records, SchemaTypeOrStr, StrDict

unique_regex = re.compile(r"Key \((?P<column>(\w+,? ?)+)\)=\((?P<value>.+)\)")


class SqlApiPath(ApiPath):
    """An :class:`.ApiPath` backed by an SQL model"""

    table: str = ""
    """sql table name"""

    @property
    def db(self) -> CrudDB:
        """Database connection pool"""
        return self.request.app["db"]

    @property
    def db_table(self) -> sa.Table:
        """Default database table for this route.

        Obtained from the :attr:`table` attribute.
        """
        return self.db.metadata.tables[self.table]

    def get_search_clause(
        self, table: sa.Table, query: Select, search: str, search_fields: Sequence[str]
    ) -> Select:
        if not search:
            return query

        columns = [getattr(table.c, col) for col in search_fields]
        return query.where(or_(*(col.ilike(f"%{search}%") for col in columns)))

    async def get_list(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ) -> PaginatedData:
        """Get a list of models

        :param filters: dictionary of filters, if not provided it will be created from
            the query_schema
        :param query: additional query parameters, only used when filters
            is not provided
        :param table: sqlalchemy table, if not provided the default :attr:`.db_table` is
            used instead
        :param conn: optional db connection
        """
        table = table if table is not None else self.db_table
        if filters is None:
            filters = self.get_filters(query=query, query_schema=query_schema)
        specials = self.get_special_params(cast(Dict, filters))
        sql_query = cast(
            Select,
            self.db.get_query(table, table.select(), params=filters, consumer=self),
        )
        #
        # order by
        if specials["order_by"]:
            order_by_column = getattr(table.c, specials["order_by"], None)
            if order_by_column is not None:
                if specials["order_desc"]:
                    order_by_column = order_by_column.desc()
                sql_query = sql_query.order_by(order_by_column)

        # search
        sql = self.get_search_clause(
            table, sql_query, specials["search"], specials["search_fields"]
        )

        # pagination
        sql_query = sql
        offset = specials["offset"]
        limit = specials["limit"]
        if offset:
            sql_query = sql_query.offset(offset)
        if limit:
            sql_query = sql_query.limit(limit)

        async with self.db.ensure_connection(conn) as conn:
            total = await self.db.db_count_query(sql, conn=conn)
            values = await conn.execute(sql_query)
        pagination = Pagination(self.full_url())
        data = cast(List[StrDict], self.dump(dump_schema, values.all()))
        return pagination.paginated(data, total, offset, limit)

    async def create_one(
        self,
        *,
        data: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        body_schema: SchemaTypeOrStr = "body_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ) -> StrDict:
        """Create a new database model

        :param data: input data, if not given it loads it via :meth:`.json_data`
        :param table: sqlalchemy table, if not given it uses the
            default :attr:`db_table`
        :param conn: optional db connection
        """
        if data is None:
            data = self.insert_data(await self.json_data(), body_schema=body_schema)
        table = table if table is not None else self.db_table
        sql = self.db.get_insert(table, data)

        async with self.db.ensure_connection(conn) as conn:
            try:
                result = await conn.execute(sql)
            except IntegrityError as exc:
                self.handle_unique_violation(exc)
                raise

        return cast(StrDict, self.dump(dump_schema, result.one()))

    async def create_list(
        self,
        *,
        data: Optional[DataType] = None,
        table: Optional[sa.Table] = None,
        body_schema: SchemaTypeOrStr = "body_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ):
        """Create multiple models"""
        table = table if table is not None else self.db_table
        if data is None:
            data = await self.json_data()
        if not isinstance(data, list):
            raise web.HTTPBadRequest(
                **self.api_response_data({"message": "Invalid JSON payload"})
            )
        schema = self.get_schema(body_schema)
        assert schema.container is list
        data = [self.insert_data(d, body_schema=schema.element) for d in data]
        values = await self.db.db_insert(table, data, conn=conn)
        return self.dump(dump_schema, values.all())

    async def get_one(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ):
        """Get a single model

        :param filters: dictionary of filters, if not provided it will be created from
            the query_schema
        :param query: additional query parameters, only used when filters
            is not provided
        :param table: sqlalchemy table, if not provided the default :attr:`.db_table` is
            used instead
        :param conn: optional db connection
        """
        table = table if table is not None else self.db_table
        if filters is None:
            filters = self.get_filters(query=query, query_schema=query_schema)
        values = await self.db.db_select(table, filters, conn=conn, consumer=self)
        row = values.first()
        if row is None:
            raise web.HTTPNotFound()
        return self.dump(dump_schema, row)

    async def update_one(
        self,
        *,
        data: Optional[StrDict] = None,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        body_schema: SchemaTypeOrStr = "body_schema",
        query_schema: SchemaTypeOrStr = "query_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ):
        """Update a single model"""
        table = table if table is not None else self.db_table
        if data is None:
            data = self.cleaned("body_schema", await self.json_data(), strict=False)
        if not filters:
            filters = self.get_filters(query=query, query_schema=query_schema)

        if data:
            try:
                values = await self.db.db_update(
                    table, filters, data, conn=conn, consumer=self
                )
            except IntegrityError as exc:
                self.handle_unique_violation(exc)
                raise
        else:
            values = await self.db.db_select(table, filters, conn=conn, consumer=self)
        row = values.first()
        if row is None:
            raise web.HTTPNotFound()
        return self.dump(dump_schema, row)

    async def delete_one(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
        conn: Optional[Connection] = None,
    ) -> Record:
        """Delete a single model"""
        table = table if table is not None else self.db_table
        if filters is None:
            filters = self.get_filters(query=query, query_schema=query_schema)
        values = await self.db.db_delete(table, filters, conn=conn, consumer=self)
        row = values.first()
        if row is None:
            raise web.HTTPNotFound()
        return row

    async def delete_list(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        conn: Optional[Connection] = None,
    ) -> Records:
        """delete multiple models"""
        table = table if table is not None else self.db_table
        if filters is None:
            filters = self.get_filters(query=query)
        return await self.db.db_delete(table, filters, conn=conn, consumer=self)

    def handle_unique_violation(self, exception: IntegrityError):
        match = re.search(unique_regex, str(exception))
        if match:
            column = match.group("column")
            message = f"{column} already exists"
            self.raise_validation_error(message=message)



import os
from typing import Dict, List, Optional

from aiohttp import web
from multidict import MultiDict

from openapi.json import dumps

MAX_PAGINATION_LIMIT: int = int(os.environ.get("MAX_PAGINATION_LIMIT") or 100)
DEF_PAGINATION_LIMIT: int = int(os.environ.get("DEF_PAGINATION_LIMIT") or 50)


class PaginatedData:
    def __init__(
        self, data: List, pagination: "Pagination", total: int, offset: int, limit: int
    ) -> None:
        self.data = data
        self.pagination = pagination
        self.total = total
        self.offset = offset
        self.limit = limit

    def json_response(self, headers: Optional[Dict[str, str]] = None, **kwargs):
        headers = headers or {}
        links = self.header_links()
        if links:
            headers["Link"] = links
        headers["X-Total-Count"] = f"{self.total}"
        kwargs.setdefault("dumps", dumps)
        return web.json_response(self.data, headers=headers, **kwargs)

    def header_links(self) -> str:
        links = self.pagination.links(self.total, self.limit, self.offset)
        return ", ".join(f'<{value}>; rel="{name}"' for name, value in links.items())


class Pagination:
    def __init__(self, url) -> None:
        self.url = url
        self.query = MultiDict(url.query)

    def paginated(
        self, data: List, total: int, offset: int, limit: int
    ) -> PaginatedData:
        return PaginatedData(data, self, total, offset, limit)

    def first_link(self, total, limit, offset):
        n = self._count_part(offset, limit, 0)
        if n:
            offset -= n * limit
        if offset > 0:
            return self.link(0, min(limit, offset))

    def prev_link(self, total, limit, offset):
        if offset:
            olimit = min(limit, offset)
            prev_offset = offset - olimit
            return self.link(prev_offset, olimit)

    def next_link(self, total, limit, offset):
        next_offset = offset + limit
        if total > next_offset:
            return self.link(next_offset, limit)

    def last_link(self, total, limit, offset):
        n = self._count_part(total, limit, offset)
        if n > 0:
            return self.link(offset + n * limit, limit)

    def link(self, offset, limit):
        query = self.query.copy()
        query.update({"offset": offset, "limit": limit})
        return self.url.with_query(query)

    def _count_part(self, total, limit, offset):
        n = (total - offset) // limit
        # make sure we account for perfect matching
        if n * limit + offset == total:
            n -= 1
        return max(0, n)

    def links(self, total, limit, offset):
        links = {}
        first = self.first_link(total, limit, offset)
        if first:
            links["first"] = first
            prev = self.prev_link(total, limit, offset)
            if prev != first:
                links["prev"] = prev
        next_ = self.next_link(total, limit, offset)
        if next_:
            last = self.last_link(total, limit, offset)
            if last != next_:
                links["next"] = next_
            links["last"] = last
        return links



import typing as t
from dataclasses import make_dataclass
from datetime import date, datetime
from decimal import Decimal

import sqlalchemy as sa
from sqlalchemy_utils import UUIDType

from . import fields

CONVERTERS = {}


def dataclass_from_table(
    name: str,
    table: sa.Table,
    *,
    exclude: t.Optional[t.Sequence[str]] = None,
    include: t.Optional[t.Sequence[str]] = None,
    required: bool = False,
    ops: t.Optional[t.Dict[str, t.Sequence[str]]] = None,
) -> type:
    """Create a dataclass from an :class:`sqlalchemy.schema.Table`

    :param name: dataclass name
    :param table: sqlalchemy table
    :param exclude: fields to exclude from the dataclass
    :param include: fields to include in the dataclass
    :param required: set all non nullable columns as required fields in the dataclass
    :param ops: additional operation for fields
    """
    columns = []
    include = set(include or table.columns.keys()) - set(exclude or ())
    column_ops = t.cast(t.Dict[str, t.Sequence[str]], ops or {})
    for col in table.columns:
        if col.name not in include:
            continue
        ctype = type(col.type)
        converter = CONVERTERS.get(ctype)
        if not converter:  # pragma:   no cover
            raise NotImplementedError(f"Cannot convert column {col.name}: {ctype}")
        field = (col.name, *converter(col, required, column_ops.get(col.name, ())))
        columns.append(field)
    return make_dataclass(name, columns)


def converter(*types):
    def _(f):
        for type_ in types:
            CONVERTERS[type_] = f
        return f

    return _


@converter(sa.Boolean)
def bl(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.bool_field)
    return (bool, data_field(**info(col, required, ops)))


@converter(sa.Integer)
def integer(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.number_field)
    return (int, data_field(precision=0, **info(col, required, ops)))


@converter(sa.Numeric)
def number(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.decimal_field)
    return (Decimal, data_field(precision=col.type.scale, **info(col, required, ops)))


@converter(sa.String, sa.Text, sa.CHAR, sa.VARCHAR)
def string(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.str_field)
    return (
        str,
        data_field(max_length=col.type.length or 0, **info(col, required, ops)),
    )


@converter(sa.DateTime)
def dt_ti(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.date_time_field)
    return (
        datetime,
        data_field(timezone=col.type.timezone, **info(col, required, ops)),
    )


@converter(sa.Date)
def dt(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.date_field)
    return (date, data_field(**info(col, required, ops)))


@converter(sa.Enum)
def en(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.enum_field)
    return (
        col.type.enum_class,
        data_field(col.type.enum_class, **info(col, required, ops)),
    )


@converter(sa.JSON)
def js(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.json_field)
    val = None
    if col.default:
        arg = col.default.arg
        val = arg() if col.default.is_callable else arg
    return (JsonTypes.get(type(val), t.Dict), data_field(**info(col, required, ops)))


@converter(UUIDType)
def uuid(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data_field = col.info.get("data_field", fields.uuid_field)
    return (str, data_field(**info(col, required, ops)))


def info(col: sa.Column, required: bool, ops: t.Sequence[str]) -> t.Tuple:
    data = dict(
        description=col.doc,
        required=not col.nullable if required is not False else False,
        ops=ops,
    )
    data.update(col.info)
    data.pop("data_field", None)
    return data


JsonTypes = {list: t.List, dict: t.Dict}
